/*******************************************************************************
*    Comm2J 1.0.                                                               *
*    Copyright (C) 2004 Luca Bertoncello                                       *
*    Hartigstrae, 12 - 01127 Dresden Deutschland                              *
*    E-Mail: lucabert@lucabert.com, lucabert@lucabert.de                       *
*    http://www.lucabert.com/  http://www.lucabert.de/                         *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

package comm2j;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import javax.microedition.io.Connection;
import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnection;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Form;

import comm2j.Message;
import comm2j.Read2J;

/**
 * Comm2j is a class used to communicate with a Server (J2ME or J2SE), transmitting
 * and receiving Messages.<br>
 * It uses Read2J to communicate with the Server. Comm2J runs <u><b>only</b></u> on J2ME.<br>
 * On J2SE you need to use just the Read2J.
 * @see MessageI
 * @see Read2J
 * @author Luca Bertoncello &lt;lucabert@lucabert.de&gt;
 * @version 1.0
*/

public class Comm2J
{
  private CommandListener cmdListener = null;
  private Display display             = null;
  private String host                 = null;
  private String port                 = null;
  private StreamConnection conn       = null;
  private int mode                    = Read2J.RAW;

  /**
   * Constructor.<br>
   * Save the connection data. It's possibile to change them using set functions
   * @param cmdListener CommandListener to be used to send commands
   * @param display Display to be used to show data
   * @param host Host to connect to
   * @param port Port on the Host to connect to
   */
  public Comm2J(CommandListener cmdListener, Display display, String host, String port)
  {
    this.cmdListener                  = cmdListener;
    this.display                      = display;
    this.host                         = host;
    this.port                         = port;
    this.mode                         = Read2J.RAW;
  }

  /**
   * Constructor.<br>
   * Save the connection data. It's possibile to change them using set functions
   * @param cmdListener CommandListener to be used to send commands
   * @param display Display to be used to show data
   * @param host Host to connect to
   * @param port Port on the Host to connect to
   * @param mode Communication mode (RAW or COMPRESSED)
   */
  public Comm2J(CommandListener cmdListener, Display display, String host, String port, int mode)
  {
    this.cmdListener                  = cmdListener;
    this.display                      = display;
    this.host                         = host;
    this.port                         = port;
    this.mode                         = mode;
  }

  /**
   * Returns the CommandListener used by Comm2J
   * @return Used CommandListener
   */
  public CommandListener getCommandListener()
  {
    return this.cmdListener;
  }

  /**
   * Sets the CommandListener used by Comm2J
   * @param cmdListener New CommandListener to be used
   */
  public void setCommandListener(CommandListener cmdListener)
  {
    this.cmdListener                  = cmdListener;
  }

  /**
   * Returns the display used by Comm2J
   * @return Current used display
   */
  public Display getDisplay()
  {
    return this.display;
  }

  /**
   * Sets the display to be used by Comm2J
   * @param display New display to be used to show data
   */
  public void setDisplay(Display display)
  {
    this.display                      = display;
  }

  /**
   * Returns the mode used by Comm2J for the communication with the server
   * @see Read2J
   * @return Used communication mode
   */
  public int getMode()
  {
    return this.mode;
  }

  /**
   * Sets the mode used by Comm2J for the communication with the server
   * @see Read2J
   * @param mode New communication mode to be used
   */
  public void setMode(int mode)
  {
    this.mode                         = mode;
  }

  /**
   * Returns the host used by Comm2j
   * @return Current host to connect to
   */
  public String getHost()
  {
    return this.host;
  }

  /**
   * Sets the host used by Comm2j
   * @param host New host to connect to
   */
  public void setHost(String host)
  {
    this.host                         = host;
  }

  /**
   * Returns the port used by Comm2j
   * @return Current port on the Host to connect to
   */
  public String getPort()
  {
    return this.port;
  }

  /**
   * Sets the port used by Comm2j
   * @param port New port on the Host to connect to
   */
  public void setPort(String port)
  {
    this.port                         = port;
  }

  /**
   * Returns the connection link to the remote host ot null if
   * not yet initialized.
   * @return Connection to the remote host
   */
  public Connection getConnection()
  {
    return this.conn;
  }

  /**
   * Connect to the remote host. Print an error on Display if
   * a connection problem occours
   */
  public void connect()
  {
    try
    {
      this.conn                       = (StreamConnection) Connector.open("socket://" + getHost() + ":" + getPort(), Connector.READ_WRITE, true);
    }
    catch(Exception e)
    {
      printErrorOnDisplay(e);
    }
  }

  /**
   * Used to know if Comm2J has a valid link to the remote host
   * @return true if valid link, false else
   */
  public boolean isConnected()
  {
    return (this.conn != null);
  }

  /**
   * Disconnect from the remote host. Print an error on Display if
   * a disconnection problem occours
   */
  public void disconnect()
  {
    try
    {
      this.conn.close();
    }
    catch(Exception e)
    {
      printErrorOnDisplay(e);
    }
  }

  /**
   * Send a message to the remote host. Print an error on Display if
   * a communication problem occours
   * @param msg Message to be sent
   * @return number of sent bytes
   */
  public int writeMessageToStream(Message msg)
  {
    try
    {
      OutputStream os                 = conn.openOutputStream();
      Read2J read2j                   = new Read2J();
      read2j.setMode(this.mode);
      return read2j.writeMessageToStream(msg, os);
    }
    catch(Exception e)
    {
      printErrorOnDisplay(e);
      return -1;
    }
  }

  /**
   * Receive a message from the remote host. Print an error on Display if
   * communication problem occours
   * @return Message read from the remote host
   */
  public Message readMessageFromStream()
  {
    try
    {
      InputStream is                  = conn.openInputStream();
      Read2J read2j                   = new Read2J();
      read2j.setMode(this.mode);
      return read2j.readMessageFromStream(is);
    }
    catch(Exception e)
    {
      printErrorOnDisplay(e);
      return null;
    }
  }

  /**
   * Show a given error on display.<br>
   * This function is just intended to be used by other functions of Comm2J class.
   * @param e Exception to be shown
   */
  private void printErrorOnDisplay(Exception exception)
  {
    printErrorOnDisplay(exception.toString());
  }

  /**
   * Show a given error on display.<br>
   * This function is just intended to be used by other functions of Comm2J class.
   * @param error Error to be shown
   */
  private void printErrorOnDisplay(String error)
  {
    Form errorF                       = new Form("Error");
    System.err.println("ERROR: [" + error + "]");
    errorF.append("ERROR: [" + error + "]");
    errorF.addCommand(new Command("Exit", Command.EXIT, 1));
    getDisplay().setCurrent(errorF);
    errorF.setCommandListener(getCommandListener());
  }
}
