/*******************************************************************************
*    Comm2J 1.0.                                                               *
*    Copyright (C) 2004 Luca Bertoncello                                       *
*    Hartigstrae, 12 - 01127 Dresden Deutschland                              *
*    E-Mail: lucabert@lucabert.com, lucabert@lucabert.de                       *
*    http://www.lucabert.com/  http://www.lucabert.de/                         *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

package comm2j;

import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Hashtable;

/**
 * LZW2J is a class used to compress/decompress an Input-/OutputStream
 * @author Luca Bertoncello &lt;lucabert@lucabert.de&gt;
 * @version 1.0
*/

public class LZW2J
{
  final static int MAX_CODES  = 1024;
  final static int BYTE_SIZE  = 8;
  final static int EXCESS     = 4;
  final static int ALPHA      = 256;
  final static int MASK       = 15;
  final static int MASK1      = 255;
  final static int MASK2      = 15;

  private int leftOver;
  private boolean bitsLeftOver;
  private byte [] s;
  private int size;
  private Element [] h;

  /**
   * Constructor.
   */
  public LZW2J()
  {
  }

  private void outputCompress(int pcode, OutputStream os) throws IOException
  {
    int c, d;

    if(this.bitsLeftOver)
    {
      d                       = pcode & MASK1;
      c                       = (this.leftOver << EXCESS) + (pcode >> BYTE_SIZE);
      os.write(c);
      os.write(d);
      this.bitsLeftOver       = false;
    }
    else
    {
      this.leftOver           = pcode & MASK2;
      c                       = pcode >> EXCESS;
      os.write(c);
      this.bitsLeftOver       = true;
    }
  }

  private void outputDecompress(int code, OutputStream os) throws IOException
  {
    this.size                 = -1;

    while(code >= ALPHA)
    {
      s[++this.size]          = (byte) h[code].getSuffix();
      code                    = h[code].getPrefix();
    }

    s[++this.size]            = (byte) code;

    for(int i = this.size; i >= 0; i--)
    {
      os.write(s[i]);
    }
  }

  private int getCode(InputStream is) throws IOException
  {
    int c                     = is.read();

    if(c == -1)
      return -1;

    int code;
    if(this.bitsLeftOver)
    {
      code                    = (this.leftOver << BYTE_SIZE) + c;
    }
    else
    {
      int d                   = is.read();
      code                    = (c << EXCESS) + (d >> EXCESS);
      this.leftOver           = d & MASK;
    }

    this.bitsLeftOver         = !this.bitsLeftOver;

    return code;
  }

  /**
   * Compress data from given InputStream and send the compressed
   * result data to given OutputStream.
   * @param is InputStream to read data from
   * @param os OutputStream to send compressed data to
   */
  public void compress(InputStream is, OutputStream os) throws IOException
  {
    Hashtable table           = new Hashtable();

    for(int i = 0; i < ALPHA; i++)
    {
      table.put(new Integer(i), new Integer(i));
    }

    int codeUsed              = ALPHA;

    int c                     = is.read();
    if(c != -1)
    {
      int pcode               = c;
      c                       = is.read();
      while(c != -1)
      {
        int k                 = (pcode << BYTE_SIZE) + c;
        Integer e             = (Integer) table.get(new Integer(k));
        if(e == null)
        {
          outputCompress(pcode, os);
          if(codeUsed < MAX_CODES)
            table.put(new Integer((pcode << BYTE_SIZE) + c), new Integer(codeUsed++));
          pcode               = c;
        }
        else
        {
          pcode               = e.intValue();
        }

        c                     = is.read();
      }

      outputCompress(pcode, os);
      if(this.bitsLeftOver)
      {
        os.write(this.leftOver << EXCESS);
      }
    }

    os.flush();
  }

  /**
   * Decompress data from given InputStream and send the uncompressed
   * result data to given OutputStream.
   * @param is InputStream to read compressed data from
   * @param os OutputStream to send uncompressed data to
   */
  public void decompress(InputStream is, OutputStream os) throws IOException
  {
    int codeUsed              = ALPHA;
    s                         = new byte[MAX_CODES];
    h                         = new Element[MAX_CODES];

    int pcode                 = getCode(is), ccode;
    if(pcode >= 0)
    {
      s[0]                    = (byte) pcode;
      os.write(s[0]);
      this.size               = 0;

      do
      {
        ccode                 = getCode(is);
        if(ccode < 0)
          break;
        if(ccode < codeUsed)
        {
          outputDecompress(ccode, os);
          if(codeUsed < MAX_CODES)
          {
            h[codeUsed++]     = new Element(pcode, (int) s[this.size]);
          }
        }
        else
        {
          h[codeUsed++]       = new Element(pcode, (int) s[this.size]);
          outputDecompress(ccode, os);
        }
        pcode                 = ccode;
      } while(true);
    }

    os.flush();
  }
}

class Element
{
  private int prefix;
  private int suffix;

  public Element(int prefix, int suffix)
  {
    this.prefix               = prefix;
    this.suffix               = suffix;
  }

  public int getPrefix()
  {
    return this.prefix;
  }

  public int getSuffix()
  {
    return this.suffix;
  }
}
